@@ -260,7 +260,7 @@ extern TC_LOG_DUMMY tc_log_dummy;
 class Relay_log_info;
 
 /*
-  Note that we destroy the lock mutex in the desctructor here.
+  Note that we destroy the lock mutex in the destructor here.
   This means that object instances cannot be destroyed/go out of scope,
   until we have reset thd->current_linfo to NULL;
  */
@@ -838,7 +838,7 @@ public:
     FD     - Format-Description event,
     R      - Rotate event
     R_f    - the fake Rotate event
-    E      - an arbirary event
+    E      - an arbitrary event
 
     The underscore indexes for any event
     `_s'   indicates the event is generated by Slave
@@ -1213,6 +1213,7 @@ public:
                         const char *user_host, size_t user_host_len, ulonglong query_utime,
                         ulonglong lock_utime, bool is_command,
                         const char *sql_text, size_t sql_text_len)= 0;
+  ATTRIBUTE_FORMAT(printf, 3, 0)
   virtual bool log_error(enum loglevel level, const char *format,
                          va_list args)= 0;
   virtual bool log_general(THD *thd, my_hrtime_t event_time, const char *user_host, size_t user_host_len, my_thread_id thread_id,
@@ -1240,6 +1241,7 @@ public:
                         const char *user_host, size_t user_host_len, ulonglong query_utime,
                         ulonglong lock_utime, bool is_command,
                         const char *sql_text, size_t sql_text_len) override;
+  ATTRIBUTE_FORMAT(printf, 3, 0)
   bool log_error(enum loglevel level, const char *format,
                          va_list args) override;
   bool log_general(THD *thd, my_hrtime_t event_time, const char *user_host, size_t user_host_len, my_thread_id thread_id,
@@ -1270,6 +1272,7 @@ public:
                         const char *user_host, size_t user_host_len, ulonglong query_utime,
                         ulonglong lock_utime, bool is_command,
                         const char *sql_text, size_t sql_text_len) override;
+  ATTRIBUTE_FORMAT(printf, 3, 0)
   bool log_error(enum loglevel level, const char *format,
                          va_list args) override;
   bool log_general(THD *thd, my_hrtime_t event_time, const char *user_host, size_t user_host_len, my_thread_id thread_id,
@@ -1326,10 +1329,12 @@ public:
   void cleanup_base();
   /* Free memory. Nothing could be logged after this function is called */
   void cleanup_end();
+  ATTRIBUTE_FORMAT(printf, 3, 0) // 1st arg is the implicit `this`
   bool error_log_print(enum loglevel level, const char *format,
                       va_list args);
   bool slow_log_print(THD *thd, const char *query, size_t query_length,
                       ulonglong current_utime);
+  ATTRIBUTE_FORMAT(printf, 4, 0)
   bool general_log_print(THD *thd,enum enum_server_command command,
                          const char *format, va_list args);
   bool general_log_write(THD *thd, enum enum_server_command command,
@@ -1367,22 +1372,26 @@ enum enum_binlog_format {
 int query_error_code(THD *thd, bool not_killed);
 uint purge_log_get_error_code(int res);
 
-int vprint_msg_to_log(enum loglevel level, const char *format, va_list args);
-void sql_print_error(const char *format, ...);
-void sql_print_warning(const char *format, ...);
-void sql_print_information(const char *format, ...);
-void sql_print_information_v(const char *format, va_list ap);
-typedef void (*sql_print_message_func)(const char *format, ...);
+int vprint_msg_to_log(enum loglevel level, const char *format, va_list args)
+  ATTRIBUTE_FORMAT(printf, 2, 0);
+void sql_print_error(const char *format, ...) ATTRIBUTE_FORMAT(printf, 1, 2);
+void sql_print_warning(const char *format, ...) ATTRIBUTE_FORMAT(printf, 1, 2);
+void sql_print_information(const char *format, ...)
+  ATTRIBUTE_FORMAT(printf, 1, 2);
+void sql_print_information_v(const char *format, va_list ap)
+  ATTRIBUTE_FORMAT(printf, 1, 0);
+typedef void (*sql_print_message_func)(const char *format, ...)
+  ATTRIBUTE_FORMAT_FPTR(printf, 1, 2);
 extern sql_print_message_func sql_print_message_handlers[];
 
 int error_log_print(enum loglevel level, const char *format,
-                    va_list args);
+                    va_list args) ATTRIBUTE_FORMAT(printf, 2, 0);
 
 bool slow_log_print(THD *thd, const char *query, uint query_length,
                     ulonglong current_utime);
 
 bool general_log_print(THD *thd, enum enum_server_command command,
-                       const char *format,...);
+                       const char *format,...) ATTRIBUTE_FORMAT(printf, 3, 4);
 
 bool general_log_write(THD *thd, enum enum_server_command command,
                        const char *query, size_t query_length);
@@ -1409,7 +1418,7 @@ binlog_cache_data* binlog_get_cache_data(binlog_cache_mngr *cache_mngr,
                                          bool use_trans_cache);
 
 extern MYSQL_PLUGIN_IMPORT MYSQL_BIN_LOG mysql_bin_log;
-extern handlerton *binlog_hton;
+extern transaction_participant binlog_tp;
 extern LOGGER logger;
 
 extern const char *log_bin_index;
@@ -1476,7 +1485,7 @@ end:
 
 static inline TC_LOG *get_tc_log_implementation()
 {
-  if (total_ha_2pc <= 1)
+  if (total_ha_2pc <= 2) // online_alter_tp and MHNSW_Trx::tp
     return &tc_log_dummy;
   if (opt_bin_log)
     return &mysql_bin_log;
@@ -1495,8 +1504,8 @@ const char *
 get_gtid_list_event(IO_CACHE *cache, Gtid_list_log_event **out_gtid_list);
 
 int binlog_commit(THD *thd, bool all, bool is_ro_1pc= false);
-int binlog_commit_by_xid(handlerton *hton, XID *xid);
-int binlog_rollback_by_xid(handlerton *hton, XID *xid);
+int binlog_commit_by_xid(XID *xid);
+int binlog_rollback_by_xid(XID *xid);
 bool write_bin_log_start_alter(THD *thd, bool& partial_alter,
                                uint64 start_alter_id, bool log_if_exists);
 #endif /* LOG_H */
