@@ -17,7 +17,7 @@
 */
 
 /*
-  Functions to autenticate and handle reqests for a connection
+  Functions to authenticate and handle requests for a connection
 */
 
 #include "mariadb.h"
@@ -305,17 +305,12 @@ end:
   started with corresponding variable that is greater then 0.
 */
 
-extern "C" uchar *get_key_conn(user_conn *buff, size_t *length,
-			      my_bool not_used __attribute__((unused)))
+extern "C" const uchar *get_key_conn(const void *buff_, size_t *length,
+                                     my_bool)
 {
+  auto buff= static_cast<const user_conn *>(buff_);
   *length= buff->len;
-  return (uchar*) buff->user;
-}
-
-
-extern "C" void free_user(struct user_conn *uc)
-{
-  my_free(uc);
+  return reinterpret_cast<const uchar *>(buff->user);
 }
 
 
@@ -324,8 +319,7 @@ void init_max_user_conn(void)
 #ifndef NO_EMBEDDED_ACCESS_CHECKS
   my_hash_init(key_memory_user_conn, &hash_user_connections,
                USER_CONN::user_host_key_charset_info_for_hash(),
-               max_connections, 0, 0, (my_hash_get_key)
-               get_key_conn, (my_hash_free_key) free_user, 0);
+               max_connections, 0, 0, get_key_conn, my_free, 0);
 #endif
 }
 
@@ -404,16 +398,12 @@ static const char *get_client_host(THD *client)
     client->security_ctx->host ? client->security_ctx->host : "";
 }
 
-extern "C" uchar *get_key_user_stats(USER_STATS *user_stats, size_t *length,
-                                     my_bool not_used __attribute__((unused)))
+extern "C" const uchar *get_key_user_stats(const void *user_stats_,
+                                           size_t *length, my_bool)
 {
+  auto user_stats= static_cast<const USER_STATS *>(user_stats_);
   *length= user_stats->user_name_length;
-  return (uchar*) user_stats->user;
-}
-
-void free_user_stats(USER_STATS* user_stats)
-{
-  my_free(user_stats);
+  return reinterpret_cast<const uchar *>(user_stats->user);
 }
 
 void init_user_stats(USER_STATS *user_stats,
@@ -480,9 +470,7 @@ void init_global_user_stats(void)
 {
   my_hash_init(PSI_INSTRUMENT_ME, &global_user_stats,
                USER_STATS::user_key_charset_info_for_hash(),
-               max_connections,
-               0, 0, (my_hash_get_key) get_key_user_stats,
-               (my_hash_free_key) free_user_stats, 0);
+               max_connections, 0, 0, get_key_user_stats, my_free, 0);
 }
 
 void init_global_client_stats(void)
@@ -490,48 +478,37 @@ void init_global_client_stats(void)
   my_hash_init(PSI_INSTRUMENT_ME, &global_client_stats,
                USER_STATS::user_key_charset_info_for_hash(),
                max_connections,
-               0, 0, (my_hash_get_key) get_key_user_stats,
-               (my_hash_free_key) free_user_stats, 0);
+               0, 0, get_key_user_stats, my_free, 0);
 }
 
-extern "C" uchar *get_key_table_stats(TABLE_STATS *table_stats, size_t *length,
-                                      my_bool not_used __attribute__((unused)))
+extern "C" const uchar *get_key_table_stats(const void *table_stats_,
+                                            size_t *length, my_bool)
 {
+  auto table_stats= static_cast<const TABLE_STATS *>(table_stats_);
   *length= table_stats->table_name_length;
-  return (uchar*) table_stats->table;
-}
-
-extern "C" void free_table_stats(TABLE_STATS* table_stats)
-{
-  my_free(table_stats);
+  return reinterpret_cast<const uchar *>(table_stats->table);
 }
 
 void init_global_table_stats(void)
 {
   my_hash_init(PSI_INSTRUMENT_ME, &global_table_stats,
-               Lex_ident_fs::charset_info(),
-               max_connections, 0, 0, (my_hash_get_key) get_key_table_stats,
-               (my_hash_free_key) free_table_stats, 0);
+               Lex_ident_fs::charset_info(), max_connections, 0, 0,
+               get_key_table_stats, my_free, 0);
 }
 
-extern "C" uchar *get_key_index_stats(INDEX_STATS *index_stats, size_t *length,
-                                     my_bool not_used __attribute__((unused)))
+extern "C" const uchar *get_key_index_stats(const void *index_stats_,
+                                            size_t *length, my_bool)
 {
+  auto index_stats= static_cast<const INDEX_STATS *>(index_stats_);
   *length= index_stats->index_name_length;
-  return (uchar*) index_stats->index;
-}
-
-extern "C" void free_index_stats(INDEX_STATS* index_stats)
-{
-  my_free(index_stats);
+  return reinterpret_cast<const uchar *>(index_stats->index);
 }
 
 void init_global_index_stats(void)
 {
   my_hash_init(PSI_INSTRUMENT_ME, &global_index_stats,
-               Lex_ident_fs::charset_info(),
-               max_connections, 0, 0, (my_hash_get_key) get_key_index_stats,
-               (my_hash_free_key) free_index_stats, 0);
+               Lex_ident_fs::charset_info(), max_connections, 0, 0,
+               get_key_index_stats, my_free, 0);
 }
 
 
@@ -823,7 +800,7 @@ bool thd_init_client_charset(THD *thd, uint cs_number)
       b. preserve non-default collations as is
 
       Perhaps eventually we should change (b) also to resolve non-default
-      collations accoding to @@character_set_collations. Clients that used to
+      collations according to @@character_set_collations. Clients that used to
       send a non-default collation ID in the handshake packet will have to set
       @@character_set_collations instead.
     */
@@ -1149,7 +1126,7 @@ void setup_connection_thread_globals(THD *thd)
 
 
 /*
-  Autenticate user, with error reporting
+  Authenticate user, with error reporting
 
   SYNOPSIS
    login_connection()
@@ -1393,7 +1370,7 @@ void do_handle_one_connection(CONNECT *connect, bool put_in_cache)
   THD *thd;
   if (!(thd= connect->create_thd(NULL)))
   {
-    connect->close_and_delete();
+    connect->close_and_delete(0);
     return;
   }
 
@@ -1418,16 +1395,6 @@ void do_handle_one_connection(CONNECT *connect, bool put_in_cache)
   thd->thr_create_utime= thr_create_utime;
   /* We need to set this because of time_out_user_resource_limits */
   thd->start_utime= thr_create_utime;
-
-  /*
-    handle_one_connection() is normally the only way a thread would
-    start and would always be on the very high end of the stack ,
-    therefore, the thread stack always starts at the address of the
-    first local variable of handle_one_connection, which is thd. We
-    need to know the start of the stack so that we could check for
-    stack overruns.
-  */
-  thd->thread_stack= (char*) &thd;
   setup_connection_thread_globals(thd);
 
   for (;;)
@@ -1468,7 +1435,7 @@ end_thread:
     if (!(connect->create_thd(thd)))
     {
       /* Out of resources. Free thread to get more resources */
-      connect->close_and_delete();
+      connect->close_and_delete(0);
       break;
     }
     delete connect;
@@ -1497,9 +1464,11 @@ end_thread:
   Close connection without error and delete the connect object
   This and close_with_error are only called if we didn't manage to
   create a new thd object.
+
+  Note: err can be 0 if unknown/not important
 */
 
-void CONNECT::close_and_delete()
+void CONNECT::close_and_delete(uint err)
 {
   DBUG_ENTER("close_and_delete");
 
@@ -1513,7 +1482,11 @@ void CONNECT::close_and_delete()
   vio_type= VIO_CLOSED;
 
   --*scheduler->connection_count;
-  statistic_increment(connection_errors_internal, &LOCK_status);
+
+  if (err == ER_CON_COUNT_ERROR)
+    statistic_increment(connection_errors_max_connection, &LOCK_status);
+  else
+    statistic_increment(connection_errors_internal, &LOCK_status);
   statistic_increment(aborted_connects,&LOCK_status);
 
   delete this;
@@ -1522,7 +1495,7 @@ void CONNECT::close_and_delete()
 
 /*
   Close a connection with a possible error to the end user
-  Alse deletes the connection object, like close_and_delete()
+  Else deletes the connection object, like close_and_delete()
 */
 
 void CONNECT::close_with_error(uint sql_errno,
@@ -1537,7 +1510,7 @@ void CONNECT::close_with_error(uint sql_errno,
     delete thd;
     set_current_thd(0);
   }
-  close_and_delete();
+  close_and_delete(close_error);
 }
 
 
