@@ -188,7 +188,8 @@ trx_purge_add_undo_to_history(const trx_t* trx, trx_undo_t*& undo, mtr_t* mtr)
   /* This function is invoked during transaction commit, which is not
   allowed to fail. If we get a corrupted undo header, we will crash here. */
   ut_a(undo_page);
-  trx_ulogf_t *undo_header= undo_page->page.frame + undo->hdr_offset;
+  const uint16_t undo_header_offset= undo->hdr_offset;
+  trx_ulogf_t *undo_header= undo_page->page.frame + undo_header_offset;
 
   ut_ad(rseg->needs_purge > trx->id);
   ut_ad(rseg->last_page_no != FIL_NULL);
@@ -265,9 +266,8 @@ trx_purge_add_undo_to_history(const trx_t* trx, trx_undo_t*& undo, mtr_t* mtr)
   than to intentionally violate ACID by committing something
   that is known to be corrupted. */
   ut_a(flst_add_first(rseg_header, TRX_RSEG + TRX_RSEG_HISTORY, undo_page,
-                      uint16_t(page_offset(undo_header) +
-                               TRX_UNDO_HISTORY_NODE), rseg->space->free_limit,
-                      mtr) == DB_SUCCESS);
+                      uint16_t(undo_header_offset + TRX_UNDO_HISTORY_NODE),
+                      rseg->space->free_limit, mtr) == DB_SUCCESS);
 
   mtr->write<2>(*undo_page, TRX_UNDO_SEG_HDR + TRX_UNDO_STATE +
                 undo_page->page.frame, undo_state);
@@ -285,8 +285,9 @@ static void trx_purge_free_segment(buf_block_t *rseg_hdr, buf_block_t *block,
   ut_ad(mtr.memo_contains_flagged(rseg_hdr, MTR_MEMO_PAGE_X_FIX));
   ut_ad(mtr.memo_contains_flagged(block, MTR_MEMO_PAGE_X_FIX));
 
-  while (!fseg_free_step_not_header(TRX_UNDO_SEG_HDR + TRX_UNDO_FSEG_HEADER +
-                                    block->page.frame, &mtr))
+  while (!fseg_free_step_not_header(block,
+				    TRX_UNDO_SEG_HDR + TRX_UNDO_FSEG_HEADER,
+				    &mtr))
   {
     rseg_hdr->fix();
     block->fix();
@@ -309,8 +310,8 @@ static void trx_purge_free_segment(buf_block_t *rseg_hdr, buf_block_t *block,
     mtr.memo_push(block, MTR_MEMO_PAGE_X_FIX);
   }
 
-  while (!fseg_free_step(TRX_UNDO_SEG_HDR + TRX_UNDO_FSEG_HEADER +
-                         block->page.frame, &mtr));
+  while (!fseg_free_step(block, TRX_UNDO_SEG_HDR + TRX_UNDO_FSEG_HEADER,
+                         &mtr));
 }
 
 void purge_sys_t::rseg_enable(trx_rseg_t &rseg)
@@ -482,6 +483,7 @@ loop:
     free_segment:
       ut_ad(rseg.curr_size >= seg_size);
       rseg.curr_size-= seg_size;
+      DBUG_EXECUTE_IF("undo_segment_leak", goto skip_purge_free;);
       trx_purge_free_segment(rseg_hdr, b, mtr);
       break;
     case TRX_UNDO_CACHED:
@@ -491,23 +493,28 @@ loop:
         if (undo->hdr_page_no == hdr_addr.page)
           goto found_cached;
       ut_ad("inconsistent undo logs" == 0);
-      if (false)
-      found_cached:
-        UT_LIST_REMOVE(rseg.undo_cached, undo);
+    found_cached:
       static_assert(FIL_NULL == 0xffffffff, "");
       if (UNIV_UNLIKELY(mach_read_from_4(TRX_RSEG + TRX_RSEG_FORMAT +
                                          rseg_hdr->page.frame)))
         trx_rseg_format_upgrade(rseg_hdr, &mtr);
-      mtr.memset(rseg_hdr, TRX_RSEG + TRX_RSEG_UNDO_SLOTS +
-                 undo->id * TRX_RSEG_SLOT_SIZE, 4, 0xff);
-      ut_free(undo);
+      DBUG_EXECUTE_IF("skip_cached_undo", goto skip_purge_free;);
+      if (UNIV_LIKELY(undo != nullptr))
+      {
+        UT_LIST_REMOVE(rseg.undo_cached, undo);
+        mtr.memset(rseg_hdr, TRX_RSEG + TRX_RSEG_UNDO_SLOTS +
+                   undo->id * TRX_RSEG_SLOT_SIZE, 4, 0xff);
+        ut_free(undo);
+      }
       mtr.write<8,mtr_t::MAYBE_NOP>(*rseg_hdr, TRX_RSEG + TRX_RSEG_MAX_TRX_ID +
                                     rseg_hdr->page.frame,
                                     trx_sys.get_max_trx_id() - 1);
       goto free_segment;
     }
   }
-
+#ifndef DBUG_OFF
+skip_purge_free:
+#endif /* !DBUG_OFF */
   hdr_addr= prev_hdr_addr;
 
   mtr.commit();
@@ -776,13 +783,16 @@ buf_block_t *purge_sys_t::get_page(page_id_t id)
 {
   ut_ad(!recv_sys.recovery_on);
 
-  buf_block_t*& undo_page= pages[id];
+  buf_block_t *&h= pages[id];
+  buf_block_t *undo_page= h;
 
   if (!undo_page)
   {
     undo_page= buf_pool.page_fix(id); // batch_cleanup() will unfix()
     if (!undo_page)
       pages.erase(id);
+    else
+      h= undo_page;
   }
 
   return undo_page;
@@ -918,7 +928,8 @@ bool purge_sys_t::choose_next_log()
         goto purge_nothing;
     }
 
-    offset= page_offset(undo_rec);
+    offset= uint16_t(undo_rec - b->page.frame);
+    ut_ad(undo_rec - b->page.frame == page_offset(undo_rec));
     tail.undo_no= trx_undo_rec_get_undo_no(undo_rec);
     page_no= id.page_no();
   }
@@ -960,12 +971,14 @@ inline trx_purge_rec_t purge_sys_t::get_next_rec(roll_ptr_t roll_ptr)
     return {nullptr, 0};
   }
 
+  buf_block_t *rec2_page= b;
   if (const trx_undo_rec_t *rec2=
       trx_undo_page_get_next_rec(b, offset, hdr_page_no, hdr_offset))
   {
   got_rec:
     ut_ad(page_no == page_id.page_no());
-    offset= page_offset(rec2);
+    ut_ad(page_offset(rec2) == rec2 - rec2_page->page.frame);
+    offset= uint16_t(rec2 - rec2_page->page.frame);
     tail.undo_no= trx_undo_rec_get_undo_no(rec2);
   }
   else if (hdr_page_no != page_no ||
@@ -981,6 +994,7 @@ inline trx_purge_rec_t purge_sys_t::get_next_rec(roll_ptr_t roll_ptr)
         rec2= trx_undo_page_get_first_rec(next_page, hdr_page_no, hdr_offset);
         if (rec2)
         {
+          rec2_page= next_page;
           page_no= next;
           goto got_rec;
         }
@@ -1046,7 +1060,7 @@ static void trx_purge_close_tables(purge_node_t *node, THD *thd)
     else if (t.second.first == reinterpret_cast<dict_table_t*>(-1));
     else
     {
-      dict_table_close(t.second.first, false, thd, t.second.second);
+      dict_table_close(t.second.first, thd, t.second.second);
       t.second.first= reinterpret_cast<dict_table_t*>(-1);
     }
   }
@@ -1169,7 +1183,10 @@ dict_table_t *purge_sys_t::close_and_reopen(table_id_t id, THD *thd,
 
   dict_table_t *table= trx_purge_table_open(id, mdl_context, mdl);
   if (table == reinterpret_cast<dict_table_t*>(-1))
+  {
+    VALGRIND_YIELD;
     goto retry;
+  }
 
   for (que_thr_t *thr= UT_LIST_GET_FIRST(purge_sys.query->thrs); thr;
        thr= UT_LIST_GET_NEXT(thrs, thr))
@@ -1184,7 +1201,8 @@ dict_table_t *purge_sys_t::close_and_reopen(table_id_t id, THD *thd,
         if (t.second.first == reinterpret_cast<dict_table_t*>(-1))
         {
           if (table)
-            dict_table_close(table, false, thd, *mdl);
+            dict_table_close(table, thd, *mdl);
+          VALGRIND_YIELD;
           goto retry;
         }
       }
